package actinScope;
/*
 * Cell.java... an arbitrary shape to hold actin growth and molecular species
 *
 * 
 */
/*  <ActinScope - for the simulation of dendritic actin networks>
    Copyright (C) <2005>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;
import javax.swing.*;
import java.lang.Math.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.universe.*;
import javax.media.j3d.*;
import javax.vecmath.*;

public class Cell extends Thing {
	static double cellXDim;	// dimensions of cell
	static double cellYDim;
	static double cellZDim;
	static double cellThickness = 0.00;		// thickness of the bacterial cell wall
	static double pistonHeight = 0.01;		// height of faux piston
	static double pistonFluc = 0;			// random fluctuation in x-position of the piston
	static double pistonPos;		// position of the piston in cell.. defined from BOTTOM of piston.. not the center
	static double cylVolume;		// volume of the cylinder
	static double sphVolume;		// volume of the 2 hemispheres
	static double cellVolume;		// entire volume contained within the rod-shaped bug
	static double microMolarChangePerMonomer;	// how much to change concentration when one monomer is used or returned
	static Color cellColor = Color.white;
	
	// for graphics
	static Box box1;
	static Box piston;
	static Transform3D pistonT3D;
	static TransformGroup pistonTG;
	static Color3f ambientC = new Color3f(1.0f,1.0f,1.0f);
	static Color3f diffuseC = new Color3f(0.9f,0.9f,0.9f);
	static Color3f specularC = new Color3f(1.0f,1.0f,1.0f);
	static Color3f emissiveC = new Color3f(1.0f,1.0f,1.0f);
	static float shiny = 128.0f;
	static Color3f pistonC = new Color3f(0.3f,0.3f,0.3f);
	
	static Material mPiston;
	static Appearance aWire;
	static Appearance aPiston;
	static boolean appearanceChanged = false;
	static boolean useWireAppearance = false;
	static boolean updatePistonGraphics = false;
	
	public Cell (Pt3D coord, double cellXDim, double cellYDim, double cellZDim) {
		super (coord);
		this.cellXDim = cellXDim;
		this.cellYDim = cellYDim;
		this.cellZDim = cellZDim;
		cellVolume = cellXDim*cellYDim*cellZDim;  // volume in cubic microns
		microMolarChangePerMonomer = 1/(1e-21*cellVolume*AvogadroNum);		//
		initializePistonPosition();
		//talkln ("Cell contains " + String.valueOf(Env.actinConc.getValue()/microMolarChangePerMonomer) + " actin monomers");
		//talkln ("              " + String.valueOf(Env.capConc.getValue()/microMolarChangePerMonomer) + " capping proteins");
		//talkln ("              " + String.valueOf(Env.arpConc.getValue()/microMolarChangePerMonomer) + " arp2/3s ");
	}
	
	public void step() {
		//figure out random fluctuation in x-position for this step
		if (Env.pistonJiggles) {
			pistonFluc = (2*Rdm.uRdm()-1)*Env.pJigAmp;
		} else {
			pistonFluc = 0;
		}
		
	}
	
	public void initializePistonPosition () {
		setPistonPosition(-cellXDim+Env.pistDist.getValue());
	}
	
	public void setPistonPosition (double pos) {
		pistonPos = pos;
		if (pistonPos > cellXDim - 2*pistonHeight) { pistonPos = cellXDim - 2*pistonHeight; }
		updatePistonGraphics = true;
	}
	
	public double getTipClearance (Pt3D endPt, Pt3D uVec) {
		if (!inCell(endPt)) { return -1e6; } 	// send back large negative number		
		double tipC = pistonPos - endPt.x + pistonFluc;
		if (tipC < 0) { 
			setPistonPosition(endPt.x); 
			return -1;							// no possible growth for this filament...impinging on piston
		}
		/*double cosineAng = Math.cos(uVec.x);	// the tip clearance considering the angle of the filament
		if (cosineAng > 1e-12) { 				
			tipC = tipC/cosineAng;				// if not parallel to piston surface (div by zero) then this is ~ tip clearance
		} else {
			tipC = 1e6;  						// if parallel then return a large number
		}*/
		return tipC;
	}
	
	public boolean inCell (Pt3D endPt) {
		if ((Math.abs(endPt.x) > cellXDim) | (Math.abs(endPt.y) > cellYDim) | (Math.abs(endPt.z) > cellZDim)) {
			return false;
		} 
		return true;
	}
	
	public double getActinConc () {
		return Env.actinConc.getValue();
	}
	
	public double getCapConc (Pt3D endPt) {
		return Env.capConc.getValue();
	}
	
	public double getCapConc () {
		return Env.capConc.getValue();
	}
	
	public double getArpAConc () {
		return Env.arpConc.getValue();
	}
	
	public boolean isActin (int numActin) {
		if (numActin*microMolarChangePerMonomer > getActinConc()) {
			return false;
		} else {
			return true;
		}
	}
	
	public void takeActin (int numActin) {
		Env.actinConc.addToValue(-numActin*microMolarChangePerMonomer);
	}
	
	public void putActin (int numActin) {
		Env.actinConc.addToValue(numActin*microMolarChangePerMonomer);
	}
	
	public boolean isCap (int numCap) {
		if (numCap*microMolarChangePerMonomer > getCapConc()) {
			return false;
		} else {
			return true;
		}
	}
	
	public void takeCap (int numCap) {
		Env.capConc.addToValue(-numCap*microMolarChangePerMonomer);
	}
	
	public void putCap (int numCap) {
		Env.capConc.addToValue(numCap*microMolarChangePerMonomer);
	}
	
	public boolean isArp (int numArp) {
		if (numArp*microMolarChangePerMonomer > getArpAConc()) {
			return false;
		} else {
			return true;
		}
	}
	
	public void takeArp (int numArp) {
		Env.arpConc.addToValue(-numArp*microMolarChangePerMonomer);
	}
	
	public void putArp (int numArp) {
		Env.arpConc.addToValue(numArp*microMolarChangePerMonomer);
	}
	
	public Pt3D rdmPtInsideCell () {
		Pt3D pt = new Pt3D((2*Rdm.uRdm()-1)*cellXDim,(2*Rdm.uRdm()-1)*cellYDim,(2*Rdm.uRdm()-1)*cellZDim);
		pt.inc(coord);
		return pt;
	}
	
	public Pt3D rdmBasalPt () {
		double xPos = coord.x - cellXDim;
		double yPos = coord.y + (2*Rdm.uRdm()-1)*cellYDim;
		double zPos = coord.z + (2*Rdm.uRdm()-1)*cellZDim;
		return new Pt3D(xPos,yPos,zPos);
	}
	
	public static void appearanceChange(boolean useWire) {
		appearanceChanged = true;
		useWireAppearance = useWire;
	}
	
	public void setWireFrameAppearance () {
		a.setLineAttributes(new LineAttributes(2.0f, LineAttributes.PATTERN_DASH_DOT,true));
		a.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_LINE,PolygonAttributes.CULL_NONE,0.0f));
	}
	
	public void setFillAppearance () {
		a.setLineAttributes(new LineAttributes(1.0f, LineAttributes.PATTERN_SOLID,true));
		a.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_FILL,PolygonAttributes.CULL_NONE,0.0f));
	}
	
	public void setGraphicsCapabilities () {
		super.setGraphicsCapabilities();
		
		aPiston.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
		aPiston.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		aPiston.setCapability(Appearance.ALLOW_POLYGON_ATTRIBUTES_WRITE);
		aPiston.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		aPiston.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
	}
	
	public void makeGraphics () {
//		 set graphics quality
		int boxTessalation = 80;
		int transMode = TransparencyAttributes.NONE;
		switch (Env.graphicsQuality) {
		case Env.LOW_QUALITY:
			transMode = TransparencyAttributes.FASTEST;
			break;
		case Env.MID_QUALITY:
			transMode = TransparencyAttributes.FASTEST;
			break;
		case Env.HIGH_QUALITY:
			transMode = TransparencyAttributes.NICEST;
			break;
		}
		aPiston = new Appearance();
		setGraphicsCapabilities();	// set capabilities for BranchGroup, TransformGroup, and Appearance
		// the main appearance
		m = new Material (Env.cellColor3f,Env.cellColor3f,Env.cellColor3f,Env.cellColor3f,shiny);
		ColoringAttributes cA = new ColoringAttributes(0.0f,0.0f,0.0f,ColoringAttributes.NICEST);
		TransparencyAttributes tA = new TransparencyAttributes ();
		tA.setTransparency(0.9f);
		tA.setTransparencyMode(transMode);
	
		a.setMaterial(m);
		a.setColoringAttributes(cA);
		a.setTransparencyAttributes(tA);
		setFillAppearance ();  // start out with the filled appearance
		
		// the piston appearance
		mPiston = new Material (pistonC,pistonC,pistonC,specularC,shiny);
		aPiston.setMaterial(mPiston);
		TransparencyAttributes tPist = new TransparencyAttributes ();
		tPist.setTransparency(0.2f);
		tPist.setTransparencyMode(transMode);
		aPiston.setTransparencyAttributes(tPist);
		
		//build the cell shape... a box for now
		float boxXDim = (float)(cellXDim);
		float boxYDim = (float)(cellYDim);
		float boxZDim = (float)(cellZDim);
		box1 = new Box (boxXDim,boxYDim,boxZDim,Primitive.GENERATE_NORMALS,a,boxTessalation);
		box1.setCapability(box1.ENABLE_APPEARANCE_MODIFY);
		Transform3D boxT3D = new Transform3D();
		TransformGroup boxTG = new TransformGroup(boxT3D);
		boxTG.addChild(box1);
		
		// build the cell piston
		piston = new Box ((float)(pistonHeight),boxYDim,boxZDim,Primitive.GENERATE_NORMALS,aPiston,boxTessalation);
		piston.setCapability(piston.ENABLE_APPEARANCE_MODIFY);
		pistonT3D = new Transform3D();
		pistonTG = new TransformGroup(boxT3D);
		pistonTG.setCapability(pistonTG.ALLOW_TRANSFORM_WRITE);
		pistonTG.addChild(piston);
		
		g3d.setTransform(t3d);
		g3d.addChild(boxTG);
		g3d.addChild(pistonTG);

		G.addChild(g3d);
		G.compile();
		graphicsMade = true;
	}
	public void updateGraphics () {
		if (updatePistonGraphics) {
			pistonT3D.setTranslation(new Vector3d((pistonPos+pistonHeight),coord.y,coord.z));
			pistonTG.setTransform(pistonT3D);
			updatePistonGraphics = false;
		}
		
	}
}
